@@ -2,8 +2,7 @@ use std::sync::Arc;
 
 use rend3::{
     format_sso,
-    resources::{CameraManager, DirectionalLightManager, InternalObject, MaterialManager, MeshBuffers},
-    types::TransparencyType,
+    resources::{CameraManager, DirectionalLightManager, MaterialManager, MeshBuffers, ObjectManager},
     ModeData,
 };
 use wgpu::{
@@ -17,9 +16,10 @@ use crate::{
     culling::{
         self,
         cpu::{CpuCuller, CpuCullerCullArgs},
-        gpu::{GpuCuller, GpuCullerCullArgs},
+        gpu::{GpuCuller, GpuCullerCullArgs, PreCulledBuffer},
         CulledObjectSet,
     },
+    material::{PbrMaterial, TransparencyType},
 };
 
 pub struct DirectionalShadowPassCullShadowsArgs<'a> {
@@ -28,13 +28,15 @@ pub struct DirectionalShadowPassCullShadowsArgs<'a> {
     pub encoder: &'a mut CommandEncoder,
 
     pub culler: ModeData<&'a CpuCuller, &'a GpuCuller>,
-    pub materials: &'a MaterialManager,
+    pub objects: &'a mut ObjectManager,
 
     pub interfaces: &'a ShaderInterfaces,
 
     pub lights: &'a DirectionalLightManager,
+
+    pub culling_input_opaque: ModeData<(), &'a mut PreCulledBuffer>,
+    pub culling_input_cutout: ModeData<(), &'a mut PreCulledBuffer>,
     pub directional_light_cameras: &'a [CameraManager],
-    pub objects: &'a [InternalObject],
 }
 
 pub struct CulledLightSet {
@@ -86,9 +88,8 @@ impl DirectionalShadowPass {
                             device: args.device,
                             camera,
                             interfaces: args.interfaces,
-                            materials: args.materials,
                             objects: args.objects,
-                            filter: |_, mat| mat.transparency == TransparencyType::Opaque,
+                            transparency: TransparencyType::Opaque,
                             sort: None,
                         }),
                         ModeData::GPU(gpu_culler) => {
@@ -98,10 +99,8 @@ impl DirectionalShadowPass {
                                 device: args.device,
                                 encoder: args.encoder,
                                 interfaces: args.interfaces,
-                                materials: args.materials,
                                 camera,
-                                objects: args.objects,
-                                filter: |_, mat| mat.transparency == TransparencyType::Opaque,
+                                input_buffer: args.culling_input_opaque.as_gpu(),
                                 sort: None,
                             });
                             args.profiler.end_scope(args.encoder);
@@ -117,9 +116,8 @@ impl DirectionalShadowPass {
                             device: args.device,
                             camera,
                             interfaces: args.interfaces,
-                            materials: args.materials,
                             objects: args.objects,
-                            filter: |_, mat| mat.transparency == TransparencyType::Cutout,
+                            transparency: TransparencyType::Cutout,
                             sort: None,
                         }),
                         ModeData::GPU(gpu_culler) => {
@@ -128,10 +126,8 @@ impl DirectionalShadowPass {
                                 device: args.device,
                                 encoder: args.encoder,
                                 interfaces: args.interfaces,
-                                materials: args.materials,
                                 camera,
-                                objects: args.objects,
-                                filter: |_, mat| mat.transparency == TransparencyType::Cutout,
+                                input_buffer: args.culling_input_cutout.as_gpu(),
                                 sort: None,
                             });
                             args.profiler.end_scope(args.encoder);
@@ -184,7 +180,7 @@ impl DirectionalShadowPass {
             match light.opaque_culled_objects.calls {
                 ModeData::CPU(ref draws) => culling::cpu::run(&mut rpass, draws, args.samplers, 0, args.materials, 2),
                 ModeData::GPU(ref data) => {
-                    rpass.set_bind_group(2, args.materials.gpu_get_bind_group(), &[]);
+                    rpass.set_bind_group(2, args.materials.get_bind_group_gpu::<PbrMaterial>(), &[]);
                     rpass.set_bind_group(3, args.texture_bg.as_gpu(), &[]);
                     culling::gpu::run(&mut rpass, data);
                 }
