@@ -1,23 +1,30 @@
+/// PBR Render Routine for rend3.
+///
+/// Contains [`PbrMaterial`] and the [`PbrRenderRoutine`] which serve as the default render routines.
+///
+/// Tries to strike a balance between photorealism and performance.
 use std::sync::Arc;
 
 use glam::Vec4;
 use rend3::{
     resources::{DirectionalLightManager, MaterialManager, TextureManager},
-    types::{TextureFormat, TextureHandle, TransparencyType},
-    util::output::OutputFrame,
-    ModeData, RenderRoutine, Renderer, RendererMode,
+    types::{TextureFormat, TextureHandle},
+    ManagerReadyOutput, ModeData, RenderRoutine, Renderer, RendererMode,
 };
 use wgpu::{
     Color, CommandBuffer, Device, LoadOp, Operations, RenderPassColorAttachment, RenderPassDepthStencilAttachment,
-    RenderPassDescriptor,
+    RenderPassDescriptor, TextureView,
 };
 
 pub use utils::*;
 
+use crate::material::{PbrMaterial, TransparencyType};
+
 pub mod common;
 pub mod culling;
 pub mod directional;
 pub mod forward;
+pub mod material;
 pub mod shaders;
 pub mod skybox;
 pub mod tonemapping;
@@ -46,6 +53,8 @@ impl PbrRenderRoutine {
         render_texture_options: RenderTextureOptions,
         output_format: TextureFormat,
     ) -> Self {
+        profiling::scope!("PbrRenderRoutine::new");
+
         let interfaces = common::interfaces::ShaderInterfaces::new(&renderer.device);
 
         let samplers = common::samplers::Samplers::new(&renderer.device, renderer.mode, &interfaces.samplers_bgl);
@@ -58,7 +67,9 @@ impl PbrRenderRoutine {
         let primary_passes = {
             let d2_textures = renderer.d2_texture_manager.read();
             let directional_light = renderer.directional_light_manager.read();
-            let materials = renderer.material_manager.read();
+            let mut materials = renderer.material_manager.write();
+            // TODO(material): figure out a better way for zero materials to work
+            materials.ensure_archetype::<PbrMaterial>(&renderer.device, renderer.mode);
             PrimaryPasses::new(PrimaryPassesNewArgs {
                 mode: renderer.mode,
                 device: &renderer.device,
@@ -101,6 +112,7 @@ impl PbrRenderRoutine {
     }
 
     pub fn resize(&mut self, renderer: &Renderer, options: RenderTextureOptions) {
+        profiling::scope!("PbrRenderRoutine::resize");
         let different_sample_count = self.render_textures.samples != options.samples;
         self.render_textures = RenderTextures::new(&renderer.device, options);
         if different_sample_count {
@@ -120,8 +132,15 @@ impl PbrRenderRoutine {
     }
 }
 
-impl RenderRoutine for PbrRenderRoutine {
-    fn render(&mut self, renderer: Arc<Renderer>, encoders: &mut Vec<CommandBuffer>, frame: &OutputFrame) {
+impl RenderRoutine<(), &TextureView> for PbrRenderRoutine {
+    fn render(
+        &mut self,
+        renderer: Arc<Renderer>,
+        cmd_bufs: flume::Sender<CommandBuffer>,
+        ready: ManagerReadyOutput,
+        _input: (),
+        output: &TextureView,
+    ) {
         profiling::scope!("PBR Render Routine");
 
         let mut encoder = renderer.device.create_command_encoder(&wgpu::CommandEncoderDescriptor {
@@ -130,38 +149,13 @@ impl RenderRoutine for PbrRenderRoutine {
 
         let mut profiler = renderer.profiler.lock();
 
-        let mut mesh_manager = renderer.mesh_manager.write();
-        let mut object_manager = renderer.object_manager.write();
-        let mut directional_light = renderer.directional_light_manager.write();
-        let mut materials = renderer.material_manager.write();
-        let mut d2_textures = renderer.d2_texture_manager.write();
-        let mut d2c_textures = renderer.d2c_texture_manager.write();
+        let mesh_manager = renderer.mesh_manager.read();
+        let directional_light = renderer.directional_light_manager.read();
+        let materials = renderer.material_manager.read();
+        let d2c_textures = renderer.d2c_texture_manager.read();
         let camera_manager = renderer.camera_manager.read();
-
-        // Do these in dependency order
-        // Level 2
-        let objects = object_manager.ready();
-
-        // Level 1
-        materials.ready(&renderer.device, &renderer.queue, &d2_textures);
-
-        // Level 0
-        let d2_texture_output = d2_textures.ready(&renderer.device);
-        let _d2c_texture_output = d2c_textures.ready(&renderer.device);
-        let directional_light_cameras = directional_light.ready(&renderer.device, &renderer.queue, &camera_manager);
-        mesh_manager.ready();
-
-        if d2_texture_output.dirty.map(|_| false, |v| v).into_common() {
-            self.primary_passes = PrimaryPasses::new(PrimaryPassesNewArgs {
-                mode: renderer.mode,
-                device: &renderer.device,
-                d2_textures: &d2_textures,
-                directional_lights: &directional_light,
-                materials: &materials,
-                interfaces: &self.interfaces,
-                samples: self.render_textures.samples,
-            });
-        }
+        // TODO(material): let this be read
+        let mut object_manager = renderer.object_manager.write();
 
         self.primary_passes.skybox_pass.update_skybox(skybox::UpdateSkyboxArgs {
             device: &renderer.device,
@@ -172,6 +166,43 @@ impl RenderRoutine for PbrRenderRoutine {
 
         let culler = self.gpu_culler.as_ref().map_cpu(|_| &self.cpu_culler);
 
+        let mut culling_input_opaque = culler.map(
+            |_| (),
+            |culler| {
+                culler.pre_cull(culling::gpu::GpuCullerPreCullArgs {
+                    device: &renderer.device,
+                    camera: &camera_manager,
+                    objects: &mut object_manager,
+                    transparency: TransparencyType::Opaque,
+                    sort: None,
+                })
+            },
+        );
+        let mut culling_input_cutout = culler.map(
+            |_| (),
+            |culler| {
+                culler.pre_cull(culling::gpu::GpuCullerPreCullArgs {
+                    device: &renderer.device,
+                    camera: &camera_manager,
+                    objects: &mut object_manager,
+                    transparency: TransparencyType::Cutout,
+                    sort: None,
+                })
+            },
+        );
+        let mut culling_input_blend = culler.map(
+            |_| (),
+            |culler| {
+                culler.pre_cull(culling::gpu::GpuCullerPreCullArgs {
+                    device: &renderer.device,
+                    camera: &camera_manager,
+                    objects: &mut object_manager,
+                    transparency: TransparencyType::Blend,
+                    sort: Some(culling::Sorting::FrontToBack),
+                })
+            },
+        );
+
         let culled_lights =
             self.primary_passes
                 .shadow_passes
@@ -180,11 +211,12 @@ impl RenderRoutine for PbrRenderRoutine {
                     profiler: &mut profiler,
                     encoder: &mut encoder,
                     culler,
-                    materials: &materials,
                     interfaces: &self.interfaces,
                     lights: &directional_light,
-                    directional_light_cameras: &directional_light_cameras,
-                    objects: &objects,
+                    directional_light_cameras: &ready.directional_light_cameras,
+                    objects: &mut object_manager,
+                    culling_input_opaque: culling_input_opaque.as_gpu_only_mut(),
+                    culling_input_cutout: culling_input_cutout.as_gpu_only_mut(),
                 });
 
         profiler.begin_scope("forward culling", &mut encoder, &renderer.device);
@@ -194,10 +226,10 @@ impl RenderRoutine for PbrRenderRoutine {
             profiler: &mut profiler,
             encoder: &mut encoder,
             culler,
-            materials: &materials,
             interfaces: &self.interfaces,
             camera: &camera_manager,
-            objects: &objects,
+            objects: &mut object_manager,
+            culling_input: culling_input_blend.as_gpu_only_mut(),
         });
 
         let cutout_culled_objects = self.primary_passes.cutout_pass.cull(forward::ForwardPassCullArgs {
@@ -205,10 +237,10 @@ impl RenderRoutine for PbrRenderRoutine {
             profiler: &mut profiler,
             encoder: &mut encoder,
             culler,
-            materials: &materials,
             interfaces: &self.interfaces,
             camera: &camera_manager,
-            objects: &objects,
+            objects: &mut object_manager,
+            culling_input: culling_input_cutout.as_gpu_only_mut(),
         });
 
         let opaque_culled_objects = self.primary_passes.opaque_pass.cull(forward::ForwardPassCullArgs {
@@ -216,15 +248,15 @@ impl RenderRoutine for PbrRenderRoutine {
             profiler: &mut profiler,
             encoder: &mut encoder,
             culler,
-            materials: &materials,
             interfaces: &self.interfaces,
             camera: &camera_manager,
-            objects: &objects,
+            objects: &mut object_manager,
+            culling_input: culling_input_opaque.as_gpu_only_mut(),
         });
 
         profiler.end_scope(&mut encoder);
 
-        let d2_texture_output_bg_ref = d2_texture_output.bg.as_ref().map(|_| (), |a| &**a);
+        let d2_texture_output_bg_ref = ready.d2_texture.bg.as_ref().map(|_| (), |a| &**a);
 
         self.primary_passes.shadow_passes.draw_culled_shadows(
             directional::DirectionalShadowPassDrawCulledShadowsArgs {
@@ -361,10 +393,10 @@ impl RenderRoutine for PbrRenderRoutine {
             interfaces: &self.interfaces,
             samplers: &self.samplers,
             source: self.render_textures.blit_source_view(),
-            target: frame.as_view(),
+            target: output,
         });
 
-        encoders.push(encoder.finish())
+        cmd_bufs.send(encoder.finish()).unwrap();
     }
 }
 
@@ -390,10 +422,12 @@ pub struct PrimaryPasses {
 }
 impl PrimaryPasses {
     pub fn new(args: PrimaryPassesNewArgs<'_>) -> Self {
+        profiling::scope!("PrimaryPasses::new");
+
         let gpu_d2_texture_bgl = args.mode.into_data(|| (), || args.d2_textures.gpu_bgl());
 
         let shadow_pipelines =
-            common::depth_pass::build_depth_pass_shader(common::depth_pass::BuildDepthPassShaderArgs {
+            common::depth_pass::build_depth_pass_pipeline(common::depth_pass::BuildDepthPassShaderArgs {
                 mode: args.mode,
                 device: args.device,
                 interfaces: args.interfaces,
@@ -403,7 +437,7 @@ impl PrimaryPasses {
                 ty: common::depth_pass::DepthPassType::Shadow,
             });
         let depth_pipelines =
-            common::depth_pass::build_depth_pass_shader(common::depth_pass::BuildDepthPassShaderArgs {
+            common::depth_pass::build_depth_pass_pipeline(common::depth_pass::BuildDepthPassShaderArgs {
                 mode: args.mode,
                 device: args.device,
                 interfaces: args.interfaces,
@@ -412,7 +446,7 @@ impl PrimaryPasses {
                 samples: args.samples,
                 ty: common::depth_pass::DepthPassType::Prepass,
             });
-        let skybox_pipeline = common::skybox_pass::build_skybox_shader(common::skybox_pass::BuildSkyboxShaderArgs {
+        let skybox_pipeline = common::skybox_pass::build_skybox_pipeline(common::skybox_pass::BuildSkyboxShaderArgs {
             mode: args.mode,
             device: args.device,
             interfaces: args.interfaces,
@@ -427,17 +461,18 @@ impl PrimaryPasses {
             materials: args.materials,
             samples: args.samples,
             transparency: TransparencyType::Opaque,
+            baking: common::forward_pass::Baking::Disabled,
         };
-        let opaque_pipeline = Arc::new(common::forward_pass::build_forward_pass_shader(
+        let opaque_pipeline = Arc::new(common::forward_pass::build_forward_pass_pipeline(
             forward_pass_args.clone(),
         ));
-        let cutout_pipeline = Arc::new(common::forward_pass::build_forward_pass_shader(
+        let cutout_pipeline = Arc::new(common::forward_pass::build_forward_pass_pipeline(
             common::forward_pass::BuildForwardPassShaderArgs {
                 transparency: TransparencyType::Cutout,
                 ..forward_pass_args.clone()
             },
         ));
-        let transparent_pipeline = Arc::new(common::forward_pass::build_forward_pass_shader(
+        let transparent_pipeline = Arc::new(common::forward_pass::build_forward_pass_pipeline(
             common::forward_pass::BuildForwardPassShaderArgs {
                 transparency: TransparencyType::Blend,
                 ..forward_pass_args.clone()
@@ -450,17 +485,17 @@ impl PrimaryPasses {
             ),
             skybox_pass: skybox::SkyboxPass::new(skybox_pipeline),
             transparent_pass: forward::ForwardPass::new(
-                Arc::clone(&depth_pipelines.opaque),
+                Some(Arc::clone(&depth_pipelines.opaque)),
                 transparent_pipeline,
                 TransparencyType::Blend,
             ),
             cutout_pass: forward::ForwardPass::new(
-                Arc::clone(&depth_pipelines.cutout),
+                Some(Arc::clone(&depth_pipelines.cutout)),
                 cutout_pipeline,
                 TransparencyType::Cutout,
             ),
             opaque_pass: forward::ForwardPass::new(
-                Arc::clone(&depth_pipelines.opaque),
+                Some(Arc::clone(&depth_pipelines.opaque)),
                 opaque_pipeline,
                 TransparencyType::Opaque,
             ),
